**   File:              JP2_XTAB.PRG
**   Authors:           Bill Ramos and Kirk Nason, Ashton-Tate
**   Modifications:     Jay Parsons, CIS 70160,340, BORBBS Jparsons
**   Date:              October 6, 1992

**   Description:       Crosstab function and related functions and
**                      procedures from the CCBOOSTR.PRG file, modified.

**   Notes:             Two asterisks "**" introduce mods by Jay Parsons (JP).
**                      Everything to do with or depending on "p_summary",
**                      the PCT, AVG, STD and VAR operations and the new files
**                      including the display file, is new.  Most everything
**                      else not commented is original code.

**  Modifications Copyright (C) 1992, Jay Parsons.

**  License is granted to all persons to use these mods for personal purposes.
**  License is granted to developers to distribute the compiled code as part
**  of dBASE or RunTime custom applications, provided that the developer
**  assumes all obligations of support.  Distribution of the source or compiled
**  code of the modifications for any consideration, whether alone or as part
** of any other product, is otherwise prohibited.

**  No warranty of merchantability, fitness for any purpose or otherwise is
**  made, and use of the modifications is at the user's sole risk.

**  The foregoing statements should not be understood to imply that the author
**  of the modifications has title to the original code or any right to permit
**  or regulate use or sale of the original code.

FUNCTION CrossTab   && UDS for Capturing Parameters for CrossTab
*-----------------------------------------------------------------------
* Capture CrossTab parameters for COM_XTAB
*
* Synopsis:
* 
* CROSSTAB( )
* 
* Description:
* 
* The CrossTab() function displays the front end to capture the 
* parameters for the COM_XTAB (compute crosstab) procedure.
* The CrossTab() function is called from within the QBE design
* surface by placing the function CROSSTAB() in any one of the field
* skeletons.  CrossTab() will not work within the file pothandle.
* 
* The user then fills in the optional title, the column to summarize
* across the top, the column to summarize along the side, the column
* to compute in the middle, and the type of operation.
** Operations supported are SUM,CNT,MAX,MIN,PCT,AVG,STD and VAR.  JP.
* 
* If the user presses F2 to perform the crosstab, the CrossTab() will
* remove the CROSSTAB() reference from the skeleton and keyboard in
* the commands to transfer to dot prompt and run EXECCROS.
* 
* If the user presses Ctrl-End to save, but not execute, the CrossTab(), 
* the CrossTab() function will save the parameters to the CROSSTAB.MEM 
* file and clear out the CROSSTAB() reference from the skeletion.
* 
* If the user presses Escape to cancel, the CrossTab() will not save
* any of the values and clear out the CROSSTAB() reference from 
* the skeletion.
* 
** The top and side fields may be any type but memo.  The summary field
** must be designated, must exist and must not be memo, even though it
** is not used for the CNT operation.  For other operations, additional
** limitations on the summary field apply:
**
** If it is a logical or memo field, only CNT and PCT is supported.
** If it is a character field, CNT, PCT, MAX and MIN are supported.
** If it is a date field, all operations but SUM are supported.
** All operations are supported for numeric and float summary fields. JP
*
* See also:
* 
* EXECCROS.PRG    Run the CrossTab and display the results
* COM_XTAB.PRG    Perform the actual CrossTab computations
* 
*-----------------------------------------------------------------------

  PRIVATE lc_talk, lc_cursor, lc_display, lc_status, lc_safety, lc_bell,;
    lc_onfirm

  SAVE SCREEN TO qbe

** changed to conform to other environmental saves. JP
  lc_talk = SET("TALK")
  SET TALK OFF

  lc_cursor = SET("CURSOR")
  SET CURSOR ON

  lc_escape = SET("ESCAPE")
  SET ESCAPE OFF

  lc_status = SET("STATUS")

  lc_safety = SET("SAFETY")
  SET SAFETY OFF

  lc_bell = SET("BELL")
  SET BELL OFF

  lc_confirm = SET("CONFIRM")
  SET CONFIRM OFF

  IF TYPE("kn_esc") = "U"
    DO key_vars
  ENDIF

** following routines shortened
  ll_iscolor = IIF( SET( "COLOR" ) = "ON", .t., .f. )
  IF TYPE("SAYCOLOR") = "U"
    PUBLIC saycolor
    saycolor = IIF( ll_iscolor, "rg+", "w/n" )
  ENDIF
  IF TYPE("FILLCOLOR") = "U"
    PUBLIC fillcolor
    fillcolor = IIF( ll_iscolor, "gb+", "n" )
  ENDIF
  IF TYPE("MESSCOLOR") = "U"
    PUBLIC messcolor
    messcolor = COLOR("MESSAGE")
    IF .NOT. "+" $ left( messcolor, AT("/", messcolor) )
      messcolor = "+" + messcolor
    ENDIF
  ENDIF

  DEFINE WINDOW Crosstab FROM 5,7 TO 18,73 DOUBLE
  DO shadowg WITH 5,7,18,73
  ACTIVATE WINDOW Crosstab

  DO XT_DisSays

  DO XT_IniVars

**  Routine to check for numeric field removed--type checking
**  has been left to the Gets procedures.  JP

  ON KEY LABEL f2       DO XT_F2Exit
  ON KEY LABEL shift-f1 DO XT_FldPick  WITH VARREAD()
  ll_escape = .F.
  ll_edit = .T.
  ll_f2 = .F.
  DO WHILE ll_edit
    DO XT_DisGets
    READ
    ln_readkey = READKEY()
    DO CASE
      CASE ln_readkey = rn_Esc
        ll_escape = .T.
        ll_edit = .F.
      CASE ln_readkey = rn_CtrlEnd
        IF ISBLANK( gc_xtop ) .OR. ISBLANK( gc_xside ) .OR. ;
             ( ISBLANK( gc_xsum ) .AND. gc_xoper # "CNT" )
          ll_edit = .T.
          ACTIVATE SCREEN
          @ 24,0
          DO Beep
          lc_msg = "The top" + IIF( gc_xoper = "CNT"," and side", ;
                ", side and summary" ) + " fields must be specified" + ;
                   " (Press SPACE)"
          @ 24, Center( lc_msg,80 ) SAY lc_msg COLOR &messcolor
          ln_tmp = INKEY(0)
          @ 24,0
          ACTIVATE WINDOW CrossTab
        ELSE
          ll_edit = ( gc_xoper = "NUL" )
        ENDIF
    ENDCASE
  ENDDO
  ON KEY LABEL f2
  ON KEY LABEL shift-f1

  lc_fullp = SET("FULLPATH")
  if .NOT. ll_escape
    SET FULLPATH ON

    gl_IsCC = XT_MakExt()               && Make EXECCROS.PRG if needed
    gc_xdbf = DBF()                     && Capture open DBF file
    gc_xtitl  = TRIM( gc_xtitl )        && trim variables
    gc_xtop   = TRIM( gc_xtop  )
    gc_xside  = TRIM( gc_xside )
    gc_xsum   = TRIM( gc_xsum  )
    gc_xoper  = TRIM( gc_xoper )
    SAVE TO crosstab ALL LIKE g?_x*     && ** changed for gl_xtotals JP

    IF ll_f2
      SET INSTRUCT OFF            && Prevent extra message when running app
      SET SAFETY OFF

      IF FILE( "xt.qbe" )         && Delete the existing XT.QBE file
        ERASE xt.qbe
      ENDIF

** Filenames shortened so only 20 characters need to be KEYBOARDed.  For some
** reason, setting TYPEAHEAD to a higher value doesn't seem to help.

      KEYBOARD "{Ctrl-Y}{Ctrl-M}"         && **Clear crosstab() from skeleton JP
      KEYBOARD "{Alt-L}S"                 && Layout menu, Save bar
      KEYBOARD "{Home}{Ctrl-Y}"           && Clear existing QBE name
      KEYBOARD "XT{Ctrl-M}"               && Save as XT.qbe
      KEYBOARD "{Ctrl-End}"               && Exit QBE
      KEYBOARD "{27}Y"                    && Exit Control Center
      KEYBOARD "DO Exx{Ctrl-M}"           && Run Exx.PRG

    ENDIF ll_f2
  ENDIF .NOT. ll_escape

  DEACTIVATE WINDOW Crosstab
  RELEASE WINDOW Crosstab
  RELEASE POPUP crosstab

  RELEASE gc_xtitl,gc_xtop,gc_xside,gc_xsum,gc_xoper,gl_xtotals
  RELEASE aa_lookup, params

  SET STATUS &lc_status.
  SET CURSOR &lc_cursor.
  SET ESCAPE &lc_escape.
  SET SAFETY &lc_safety.
  SET TALK &lc_talk.
  SET FULLPATH &lc_fullp
  SET CONFIRM &lc_confirm
  SET BELL &lc_bell

  RESTORE SCREEN FROM qbe
  RELEASE SCREEN qbe
RETURN ""
*-- EOF: CrossTab()

PROCEDURE XT_F2Exit  && UDS - CrossTab()
*-----------------------------------------------------------------------
* Handle the ON KEY LABEL f2 by setting the state variable LL_F2 to
* true and keyboarding a Ctrl-End into the CrossTab form.
*-----------------------------------------------------------------------
  ll_f2 = .T.
  KEYBOARD CHR( kn_CtrlEnd )
RETURN
*-- EOP: XT_F2Exit


PROCEDURE XT_DisSays && UDS - CrossTab()
*-----------------------------------------------------------------------
* Display the form with the appropiate colors.  Uses the PUBLIC
* variable FILLCOLOR to display in inside fill color.  Uses the PUBLIC
* variable SAYCOLOR to display the text information and lines.
*-----------------------------------------------------------------------
  @ 0,0 SAY REPLICATE( CHR(219), 65)
  @ 1,0 SAY REPLICATE( CHR(219), 65)
  @ 2,0 SAY REPLICATE( CHR(219), 65)
  @ 3,0 SAY REPLICATE( CHR(219), 65)
  @ 4,0 SAY REPLICATE( CHR(219), 65)
  @ 5,0 SAY REPLICATE( CHR(219), 65)
  @ 6,0 SAY REPLICATE( CHR(219), 65)
  @ 7,0 SAY REPLICATE( CHR(219), 65)
  @ 8,0 SAY REPLICATE( CHR(219), 65)
  @ 9,0 SAY REPLICATE( CHR(219), 65)
  @ 10,0 SAY REPLICATE( CHR(219), 65)
  @ 11,0 SAY REPLICATE( CHR(219), 65)
  @ 0,0 FILL TO 11,64 COLOR &fillcolor

*-- @ SAY Processing. -----------------------------------------------------
  @ 0,24 SAY "Cross Tabulation" COLOR &saycolor
  @ 1,0 SAY "" COLOR &saycolor
  @ 2,10 SAY "Title:" COLOR &saycolor
  @ 4,18 SAY "Top Field" COLOR &saycolor
  @ 4,28 SAY ">" COLOR &saycolor
  @ 5,11 SAY "" COLOR &saycolor
  @ 6,0 SAY "Side Field:" COLOR &saycolor
  @ 7,5 SAY "" COLOR &saycolor
  @ 7,11 SAY "" COLOR &saycolor
  @ 8,5 SAY "V" COLOR &saycolor
  @ 8,11 SAY "" COLOR &saycolor
  ** rearranged
  @ 8,15 SAY "Summary Field:" COLOR &saycolor
  @ 8,32 SAY "Operation:" COLOR &saycolor
  ** added for summaries
  @ 8,45 SAY "Summarize?" COLOR &saycolor
  @ 9,11 SAY "" COLOR &saycolor
  @ 10,11 SAY "" COLOR &saycolor
  @ 11,11 SAY "" COLOR &saycolor

RETURN               
*-- EOP: XT_DisSays

PROCEDURE XT_DisGets && UDS - CrossTab()
*--------------------------------------------------------------------
* Display the GET fields for the form.  Display the Navigation line
* before starting the GET processing.  Allow Shift-F1 picklists
* for the field selections.
*--------------------------------------------------------------------

*-- Perform the GETs
  @  2, 17 GET gc_xtitl  PICTURE "@S40";
    MESSAGE "Specify an optional heading for the CrossTab display";
    WHEN XT_NavDsp( 1 )

  @  4, 34 GET gc_xtop   PICTURE "@!S10" ;
    MESSAGE "Specify the field to summarize along the top of the CrossTab";
    WHEN XT_NavDsp( 2 );
    VALID REQUIRED IsInView( gc_xtop ) .AND. TYPE( gc_xtop )<>"M";
    ERROR IIF( TYPE(gc_xtop)="M",;
                  "The top field cannot be a Memo field",;
                  "The field specified is not in the current view")

  @  9,  0 GET gc_xside  PICTURE "@!S10" ;
    MESSAGE "Specify the field to summarize along the side of the CrossTab";
    WHEN XT_NavDsp( 3 );
    VALID REQUIRED IsInView( gc_xside ) .AND. TYPE( gc_xside )<>"M";
    ERROR IIF( TYPE(gc_xside)="M",;
                  "The side field cannot be a Memo field",;
                  "The field specified is not in the current view")

  @  9, 19 GET gc_xsum   PICTURE "@!S10" ;
    MESSAGE "Specify the field to operate on in the body of the CrossTab";
    WHEN XT_NavDsp( 4 );
    VALID REQUIRED IsinView( gc_xsum ) ;
    ERROR "The field specified is not in the current view"

** Changed JP
  @  9, 39 GET gc_xoper  PICTURE "@M CNT,SUM,MAX,MIN,PCT,AVG,STD,VAR,NUL";
    MESSAGE "Specify the operation on the summary field (or NUL)";
    WHEN XT_NavDsp( 5 );
    VALID REQUIRED XT_Goodop( type( gc_xsum ), gc_xoper );
    ERROR gc_Opmsg

** added JP
  @  9, 54 GET gl_xtotals PICTURE "Y" ;
     MESSAGE "Include summary row and column? Y/N" ;
     WHEN XT_NavDsp( 6 )

RETURN
*-- EOP: XT_DisGets

FUNCTION XT_Goodop
PARAMETERS lc_xsum, lc_xoper
**--------------------------------------------------------------------
** Test the chosen operation against summary field and supply message
**    added by Jay Parsons
**--------------------------------------------------------------------
  DO CASE
    CASE lc_xoper = "NUL"
      RETURN .T.
    CASE lc_xsum $ "ML"
      gc_Opmsg = "Only CNT and PCT are allowed for Memo and Logical fields"
      RETURN lc_xoper $ "CNT,PCT"
    CASE lc_xsum = "C"
      gc_Opmsg = "SUM, AVG, STD and VAR are not allowed for Character fields"
      RETURN lc_xoper $ "CNT,MAX,MIN,PCT"
    CASE lc_xsum = "D"
      gc_Opmsg = "SUM is not allowed for Date fields"
      RETURN lc_xoper # "SUM"
    OTHERWISE
      RETURN .T.
  ENDCASE
*-- EOP: XT_Goodop

PROCEDURE Com_XTab  && Compute CrossTab Based on Parameters
PARAMETERS p_topfld, p_sidfld, p_calfld, p_calc, p_summary
*------------------------------------------------------------------
* Compute data normalized form of a relational database 
* file to a spreadsheet-like structure.  It summarizes the 
* data in one field by expressing it in terms of two other 
* fields.
* 
* SYNTAX 
* 
** summary row/col added JP
* DO Com_XTab WITH <top field name>, <side field name>, < calc field >, ;
*                  <type calc> [, <include summary row/col> ]
* 
* where <type calc> can be one of the following functions:
* 
*   CNT
*   MAX
*   MIN
*   SUM
**      Functions added by JP
**  PCT        && percentage of the counted population
**  AVG        && average
**  STD        && population standard deviation
**  VAR        && population variance
* 
* USAGE
* 
* The Com_XTab command handles all records in the active 
* file that match any filter condition placed on the data.
* Com_XTab places the result of the CROSSTAB in a file
* named CROSSTAB.  Com_XTab will also place the column names
* for the <top field name> in a file named COLUMNS.  
* Com_XTab will overwrite these files.
**
**  Comments added by JP
**
**  In these days of inexpensive hard disks, the program has been
**  revised to create additional files:
**
**  The file XT_COUNT is used by all operations to hold the count
**  of items for each cell--that is, matching the values of the top
**  and side fields that intersect at that row and column.  If the
**  desired operation is CNT or PCT, this file is simply copied to CROSSTAB
**  on completion.  Although this file is not mathematically needed
**  for operations other than AVG, STD and VAR, its inclusion makes it
**  possible to create XT_DISP with true blank fields rather than
**  ambiguous values.
**
**  For other operations, the file CROSSTAB is used to hold the contents
**  of the cells required by the operation--the MAX, MIN or SUM of
**  the items that match that cell's combination of top and side values.
**  For the operations AVG, STD and VAR, the sums in CROSSTAB are then
**  replaced with those statistics.
**
**  For the operations STD and VAR, the file XT_SQRS is used to hold the
**  sums of the squares of the items in the respective CROSSTAB cells.
**  This figure is used, with the square of the sums for the cells from
**  CROSSTAB and the counts from XT_COUNT, to calculate the statistics,
**  which are placed in CROSSTAB.
**
**  In all cases, the CROSSTAB file is translated to XT_DISP for display.
**  XT_DISP uses all character fields and is thus able to show cells
**  that have no data as blanks rather than ambiguous zeroes.  While
**  not strictly necessary if the cell contents are dates, this is done
**  anyway for consistency.
**
**  The files CROSSTAB and XT_COUNT are left, as are all the files created
**  except the structure file, so that further operations may be performed
**  on them as desired.
**
**  No provision is made for avoiding overwriting of existing files with
**  the same names.  Users desiring to avoid overwrites may wish to obtain
**  a Tempname() function, such as the one in Libxx.zip on BORBBS, and
**  use it to assign the actual filenames, then saving them and their
**  aliases to a .MEM file.
*
* OPTIONS
* 
* CNT counts the records that match the <top field name>
* and the <side field name>.  ** It does not use the <calc field> data. JP
* 
* SUM adds together the values of <calc field> that match
* the <top field name> and the <side field name>.
* 
* MAX determines the maximum value of <calc field> for
* matching values of <top field name> and <side field name>.
* 
* MIN determines the minimum value of <calc field> for
* matching values of <top field name> and <side field name>.
* 
** Comments added by JP
**
** PCT divides each item of the CNT by the total number of items
** and multiplies each by 100.
**
** AVG averages the values of <calc field> that match the values
** of <top field name> and <side field name>.
**
** STD calculates the population standard deviation of the values
** of <calc field> that match the <top field name> and <side field
** name>, the square root of VAR.
**
** VAR calculates the population variance of the values of <calc field>
** that match the <top field name> and <side field name>.
**
** For sample STD and VAR calculations instead of the population measures
** provided, change the logical variable "ll_sample" near the top of the code.
**
* EXAMPLE
* 
* Consider the following database file.
* 
* CAMPING   Name  Item     Price
* -------   ----  -----   ------
*     1     John  Tent    245.00
*     2     Mark  Stove    59.00
*     3     Jane  Heater   75.00
*     4     Mark  Stove    79.00
*     5     Jane  Heater   95.00
*     6     Jane  Tent    325.00
*     7     John  Tent    175.00
*     8     Mark  Stove    89.00
* 
* DO Com_XTab WITH "Item", "Name", "Price", "SUM", .F.
* 
*   will create the following two files:
* 
** example modified to show new column names, change to avoid use
** of numbered work areas. JP
* CROSSTAB  Name  COL002  COL003  COL004
* --------  ----  ------  ------  ------
*     1     Jane  170.00    0.00  325.00  
*     2     John    0.00    0.00  420.00
*     3     Mark    0.00  227.00    0.00
* 
* The CROSSTAB file will be active in work area alias "Crosstab".
*
* COLUMNS   Column  <TopField>
* -------   ------  --------
*     1             Name
*     2     COL002  Heater
*     3     COL003  Stove
*     4     COL004  Tent
* 
* The COLUMNS file will be open in work area alias "Columns".  The COLUMNS
* file will have an MDX TAG Column on the Column field.  
* 
**
** Comments by JP
** If gl_xtotals = .T., the files will also contain a field and a
** record summarizing the data in that row and column.
** The XT_COUNT, XT_SQRS and XT_DISP files share the basic structure of
** Crosstab and map directly to it by row and column; they in effect
** add a third dimension to hold the values needed.
**
* COM_XTAB will create a SET FIELDS TO command for each XT_DISP column
* based on the Columns file to get the correct column heading.  This
* technique will only work if the resulting column name is less than or
* equal to 10 characters and it contains no punctuation or spaces.
** The XT_ClnTl function works around this. JP
* 
* You may use the CROSSTAB or XT_DISP file in the report designer to
* place more meaningful labels on the columns of a report.  Before entering
* the report designer execute the command:
*     SET FIELDS TO <Enter>
* 
*------------------------------------------------------------------

** Programmer-changeable variables, placed here for convenience

  PRIVATE ln_lblmin, ld_stdate, ll_sample
  ln_lblmin = 7                 && minimum field width for label column
                                && to suit the word "Summary" at bottom
  ld_stdate = {12/31/1899}      && base date for date arithmetic
  ll_sample = .F.               && make .T. for sample STD and VAR

** Environment saving

  PRIVATE lc_talk               && **slightly changed by JP
  lc_talk = SET("TALK")
  SET TALK OFF

  PRIVATE lc_safety
  lc_safety = SET("SAFETY")     && Save the SAFETY state
  SET SAFETY OFF                && and turn it off.

  PRIVATE lc_excl
  lc_excl = SET("EXCLUSIVE")    && Save the EXCLUSIVE state
  SET EXCLUSIVE ON              && and turn it on.

  PRIVATE lc_status
  lc_status = SET("STATUS")     && Save the status bar before turning it off
  SET STATUS OFF

** Set up values to show progress and issue first report
** changed to array to simplify conditional printing

  PRIVATE lc_stat,lc_disprow
  DECLARE lc_stat[7]
  lc_stat[ 1 ] = "Determining the number of UNIQUE Records"
  lc_stat[ 2 ] = "Creating structures and files needed    "
  lc_stat[ 3 ] = "Populating the COLUMNS file             "
  lc_stat[ 4 ] = "Populating the CROSSTAB file            "
  lc_stat[ 5 ] = "Calculating summary row and column      "
  lc_stat[ 6 ] = "Creating displayable file               "
  lc_stat[ 7 ] = "Populating the displayable file         "
  IF p_summary
    DEFINE WINDOW com_xtab FROM 5, 9 TO 16, 70
    DO ShadowG WITH 5, 9, 16, 70
  ELSE
    DEFINE WINDOW com_xtab FROM 5, 9 TO 15, 70
    DO ShadowG WITH 5, 9, 15, 70
  ENDIF
  ACTIVATE WINDOW com_xtab

  @ 0, 20 SAY "Computing CrossTab"
  ln_disprow = 1
  DO XT_BusyMsg WITH 1

** Verify possibility of continuing

*-- Make sure a data file is open
  Mainfile = ALIAS()
  IF ISBLANK( Mainfile )
    lc_error = XT_ErrMsg( "No file is in use!", "" )
    DO XT_RestEv
    CANCEL
  ENDIF   

*-- Preserve the order of the file going into crosstab
  lc_order = ORDER()
  IF ISBLANK( lc_order )
    lc_for = ""
  ELSE
    lc_for = "FOR " + FOR()
  ENDIF
  lc_filter = set( "FILTER" )
  IF .NOT. ISBLANK( lc_filter )
    lc_for = IIF( ISBLANK( lc_for ), "FOR " + lc_filter, ;
              lc_for +" .AND. " + lc_filter )
  ENDIF
  SET FILTER TO
  SET ORDER TO

** Following 3 routines changed slightly. JP
*-- Strip the aliases from the names of the top, side and calculated
*-- fields if they exist, and place in upper case
  lc_ptopfld = UPPER( p_topfld )
  IF "->" $ lc_ptopfld
    lc_ptopfld = SUBSTR( lc_ptopfld, AT( "->", lc_ptopfld + 2 ) )
  ENDIF
  lc_psidfld = UPPER( p_sidfld )
  IF "->" $ lc_psidfld
    lc_psidfld = SUBSTR( lc_psidfld, AT( "->", lc_psidfld + 2 ) )
  ENDIF
  lc_pcalfld = UPPER( p_calfld )
  IF "->" $ lc_pcalfld
    lc_pcalfld = SUBSTR( lc_pcalfld, AT( "->", lc_pcalfld + 2 ) )
  ENDIF

*-- Verify the calculation type ** changed to use Goodop JP
  IF .NOT. XT_Goodop( lc_pcalfld, p_calc ) .OR. p_calc = "NUL"
    lc_error = XT_ErrMsg( "Calculation type " + p_calc + " for " ;
                        + lc_pcalfld +" is not supported!", "" )
    DO XT_RestEv
    CANCEL
  ENDIF

*------------------------------------------------------------------
*  Having completed preliminary environment saves and checks,
*  establish dimensions, field sizes and types and the like for the
*  cross-tabulation.
*------------------------------------------------------------------

*-- Compute the number of top values
  lc_typetop = TYPE( lc_ptopfld )               && type of the top field
  DO CASE
    CASE lc_typetop = "L"
      INDEX ON IIF( &lc_ptopfld., "T", "F" ) TAG topfld UNIQUE &lc_for
    CASE lc_typetop = "C"
      INDEX ON UPPER( &lc_ptopfld. ) TAG topfld UNIQUE &lc_for
    CASE lc_typetop $ "NF"
      INDEX ON &lc_ptopfld. TAG topfld UNIQUE &lc_for
    CASE lc_typetop = "D"
      INDEX ON DTOS( &lc_ptopfld. ) TAG topfld UNIQUE &lc_for
  ENDCASE

  COUNT TO ln_columns           && unique values taken on by top field,
                                && which is columns of data for crosstab

** changed for p_summary, JP
  IF ln_columns > IIF( p_summary, 253, 254 )
    lc_error = XT_ErrMsg( "Unique columns for the top exceeds 254", "" )
    DELETE TAG topfld
    DO XT_RestEv
    CANCEL
  ENDIF

** and side values
  lc_typesid = TYPE( lc_psidfld )
  DO CASE
    CASE lc_typesid = "L"
      INDEX ON IIF( &lc_psidfld., "T", "F" ) TAG sidfld UNIQUE &lc_for
    CASE lc_typesid = "C"
      INDEX ON UPPER( &lc_psidfld. ) TAG sidfld UNIQUE &lc_for
    CASE lc_typesid $ "NF"
      INDEX ON &lc_psidfld. TAG sidfld UNIQUE &lc_for
    CASE lc_typesid = "D"
      INDEX ON DTOS( &lc_psidfld. ) TAG sidfld UNIQUE &lc_for
  ENDCASE

  COUNT TO ln_rows              && unique values of side field,
                                && which is rows of crosstab data

  DO XT_BusyMsg WITH 2

*-- Create the COLUMNS file to hold the top field values
  COPY TO _cross STRUCTURE EXTENDED
  SELECT SELECT()
  USE _cross        

*-- Get the attributes for the side field from the structure file
  LOCATE FOR field_name = lc_psidfld
  ln_lensid = field_len
  ln_decsid = field_dec

*-- Get the attributes for the top field ditto
  LOCATE FOR field_name = lc_ptopfld
  ln_lentop = field_len
  ln_dectop = field_dec

*-- Get the attributes for the calculated column
** Also figure width needed for calculated cells of the various files
** and determine data type of calculated cells, which may not be the same
** as the type of data in the calc field
** after the underscore, "x" = crosstab file ( and later its totaling array)
**                       "c" = count file            ditto
**                       "q" = squares file          ditto

  LOCATE FOR field_name = lc_pcalfld
  lc_typecal  = field_type                      && data type of calc field
  ln_cwid = INT( LOG10( RECCOUNT( mainfile ) ) ) + 1  && width for # records
  ln_cdec = 0                                   && count file needs no decimals
  lc_xtype = "N"                                && usually, xtab is numeric

  DO CASE
    CASE p_calc = "CNT"
      ln_xwid = ln_cwid                         && count & xtab files same
      ln_xdec = ln_cdec                         && no decimals needed
    CASE p_calc = "PCT"
      ln_xdec = MIN( MAX( field_dec, SET( "DECIMALS" ) ), 16 )
      ln_xwid = ln_xdec + 4                     && plus four more for "1xx."
    CASE p_calc $ "MAX,MIN"
      ln_xwid = field_len                       && no sums needed
      lc_xtype = lc_typecal                     && all data types OK
      ln_xdec = field_dec
    OTHERWISE                                   && SUM,AVG,STD,VAR
      ln_xwid = MIN( field_len + ln_cwid, 20 )  && sums may need this many
      ln_xdec = MAX( 0, field_dec - MAX( 0, field_len + ln_cwid - 20 ) )
      ln_qwid = MIN( field_len + field_len + ln_cwid, 20 )
      ln_qdec = MAX( 0, field_dec - MAX( 0, field_len + field_len ;
         + ln_cwid - 20 ) )
  ENDCASE

*----------------------------------------------------------------------
*-- Create records in the structure file for the COLUMN file fields
*----------------------------------------------------------------------
  ZAP
  APPEND BLANK
  REPLACE field_name WITH "COLUMN", ;
          field_type WITH "C"     , ;
          field_len  WITH 6       , ;
          field_dec  WITH 0       , ;
          field_idx  WITH "Y"
  APPEND BLANK            
  REPLACE field_name WITH lc_ptopfld, ;
          field_type WITH "C"       , ;
          field_len  WITH ln_lentop , ;
          field_dec  WITH 0         , ;
          field_idx  WITH "N"
                       
  CREATE columns FROM _cross
  USE columns ALIAS columns
  INDEX ON UPPER( &lc_ptopfld ) TAG &lc_ptopfld

*----------------------------------------------------------------------
*-- Create records in the structure file for the crosstab file fields
*----------------------------------------------------------------------
*-- Make the column for side field values
  SELECT select()
  USE _cross
  ZAP

  APPEND BLANK
**  Changed to type "C", min length ln_lblmin - easier to deal with labels JP
  REPLACE field_name WITH lc_psidfld, ;
          field_type WITH "C", ;
          field_len  WITH max( ln_lensid, ln_lblmin ) , ;
          field_dec  WITH 0 , ;
          field_idx  WITH "N"

*-- Make the crosstab cell columns ** Renumbered and type revised, JP
  ln_col = 2
  DO WHILE ln_col <= ln_columns + 1
    lc_col = "COL" + RIGHT( STR( 1000 + ln_col ), 3 )
    APPEND BLANK
    REPLACE field_name WITH lc_col  , ;
            field_type WITH lc_xtype, ;
            field_len  WITH ln_xwid , ;
            field_dec  WITH ln_xdec , ;
            field_idx  WITH "N"

    ln_col = ln_col + 1
  ENDDO

**-- Make the summary column.  Added by JP
  IF p_summary
    APPEND BLANK
    REPLACE field_name WITH "SUMMARY", ;
            field_type WITH lc_xtype , ;
            field_len  WITH ln_xwid  , ;
            field_dec  WITH ln_xdec  , ;
            field_idx  WITH "N"
  ENDIF

  CREATE crosstab FROM _cross
  USE crosstab ALIAS crosstab
  INDEX ON UPPER( &lc_psidfld ) TAG &lc_psidfld

*----------------------------------------------------------------------
*-- Modify records in the structure file for the COUNT and SQRS fields
*----------------------------------------------------------------------

  SELECT select()
  USE _cross
  GO 2                                  && 1st record, the name, is OK

  SCAN WHILE .NOT. eof()
    REPLACE field_type WITH "N"       , ;
            field_len  WITH ln_cwid   , ;
            field_dec  WITH ln_cdec   , ;
            field_idx  WITH "N"

  ENDSCAN

  CREATE XT_count FROM _cross
  USE XT_count ALIAS counts

  IF p_calc $ "STD,VAR"
    SELECT select()
    USE _cross
    GO 2                                && 1st record, the name, is OK

    SCAN WHILE .NOT. eof()
      REPLACE field_type WITH "N"       , ;
              field_len  WITH ln_qwid   , ;
              field_dec  WITH ln_qdec   , ;
              field_idx  WITH "N"

    ENDSCAN

    CREATE XT_sqrs FROM _cross
    USE XT_sqrs ALIAS Squares
  ENDIF

  DO XT_BusyMsg WITH 3

**--------------------------------------------------------------------
** The files needed have now been created with appropriate structures,
** but with no data.
**--------------------------------------------------------------------

** Fill in the records of the column file.  Each has one field with
** the name of the column (field) in the other files, and one field
** with one of the values appearing in the top field of the main file
** obtained from its UNIQUE index.  Each of these values is converted
** to character type to simplify the coding of later lookups.

** one for each unique value found in the main file top fields
  SELECT ( mainfile )
  SET ORDER TO TAG topfld
  ln_col = 2                    && because col 1 of crosstab is sidefld names

  SCAN      
    lc_col =  "COL" + RIGHT( STR( 1000 + ln_col ) , 3 )
    DO CASE
      CASE lc_typetop = "C"
        lc_field = &p_topfld
      CASE lc_typetop = "L"
        lc_field = IIF( &p_topfld., "T", "F" )
      CASE lc_typetop $ "NF"
        lc_field = STR( &p_topfld, ln_lentop, ln_dectop )
      CASE lc_typetop = "D"
        lc_field = DTOS( &p_topfld. )
    ENDCASE
    SELECT columns
    APPEND BLANK
    REPLACE column WITH lc_col, &lc_ptopfld WITH lc_field
    ln_col = ln_col + 1
    SELECT ( mainfile )
  ENDSCAN
                                                         
*-- Initialize the various matrices by filling in the side column of each
*-- with the UNIQUE values from the source DBF, in order.

  SET ORDER TO TAG sidfld
  IF p_summary
    ln_ordno = 1                        && this is record order
  ENDIF
  SCAN
    DO CASE
      CASE lc_typesid = "C"
        lc_field = &p_sidfld
      CASE lc_typesid = "L"
        lc_field = IIF( &p_sidfld., "T", "F" )
      CASE lc_typesid $ "NF"
        lc_field = STR( &p_sidfld, ln_lentop, ln_dectop )
      CASE lc_typesid = "D"
        lc_field = DTOS( &p_sidfld. )
    ENDCASE
    SELECT crosstab
    APPEND BLANK
    REPLACE &lc_psidfld WITH lc_field
    SELECT counts
    APPEND BLANK
    REPLACE &lc_psidfld WITH lc_field
    IF p_calc $ "STD,VAR"
      SELECT squares
      APPEND BLANK
      REPLACE &lc_psidfld WITH lc_field
    ENDIF
    SELECT ( mainfile )
  ENDSCAN

  IF p_summary
    ** add a blank record for summary row, last in order JP
    lc_botrow = "Z" + replicate( "_", ln_lensid - 1 )
    SELECT crosstab
    APPEND BLANK
    REPLACE &lc_psidfld WITH lc_botrow
    SELECT counts
    APPEND BLANK
    REPLACE &lc_psidfld WITH lc_botrow
  ENDIF

  DO XT_BusyMsg WITH 4

**----------------------------------------------------------------------
** So far, all we have done is create the files needed for the cross-
** tabulation.  We now go through the file being cross-tabulated using
** its data, as required, in the appropriate cells of the matrices.
**----------------------------------------------------------------------

*-- Populate the matrix - brute force

*-- Remove the UNIQUE Tags on the data file and restore order and filter
  SELECT ( mainfile )
  DELETE TAG topfld
  DELETE TAG sidfld
  IF .NOT. ISBLANK( lc_order )
    SET ORDER TO &lc_order
  ENDIF
  SET FILTER TO &lc_filter

  ln_pop = 0                      && reset a counter for PCT population

  SCAN                            && go through the entire main file

    lx_val = &lc_pcalfld          && this is the value for the matrix
                                                        
** Use the value found in the topfield of the record, converted to char,
** to identify the record in Columns, and look up in it the name of the
** column in the crosstab files.  Substantially changed by JP

    DO CASE
      CASE lc_typetop = "C"
        l_seekval = upper( &lc_ptopfld )
      CASE lc_typetop = "L"
        l_seekval = IIF( &lc_ptopfld., "T", "F" )
      CASE lc_typetop $ "NF"
        l_seekval = STR( &lc_ptopfld, ln_lentop, ln_dectop )
      CASE lc_typetop = "D"
        l_seekval = DTOS( &lc_ptopfld )
    ENDCASE

    IF SEEK( l_seekval, "columns" )
      lc_col = columns->column
    ELSE
      lc_error = XT_ErrMsg( "Data in the working files is corrupted.", "" )
      DO XT_RestEv
      CANCEL
    ENDIF

**  Pick up the value of the side field.  Type changed to character JP
    DO CASE
      CASE lc_typesid = "C"
        l_seekval = UPPER( &lc_psidfld )
      CASE lc_typesid = "L"
        l_seekval = IIF( &lc_psidfld., "T", "F" )
      CASE lc_typesid $ "NF"
        l_seekval = STR( &lc_psidfld., ln_lensid, ln_decsid )
      CASE lc_typesid = "D"
        l_seekval = DTOS( &p_sidfld )
    ENDCASE

**  and use it to position the file pointer in crosstab
    IF .NOT. seek( l_seekval, "crosstab" )
      lc_error = XT_ErrMsg( "Data in the working files is corrupted.", "" )
      DO XT_RestEv
      CANCEL
    ENDIF

** use the pointer to adjust the cell in Counts, and add to the population
    SELECT counts
    GO recno( "crosstab" )
    ln_count = &lc_col + 1
    REPLACE &lc_col WITH ln_count
    ln_pop = ln_pop + 1

** and then the cells in Crosstab and maybe Squares
    IF .NOT. p_calc $ "CNT,PCT"
      SELECT crosstab
      IF p_calc $ "SUM,AVG,STD,VAR"
        IF lc_typecal = "D"                    && adjust dates for arithmetic
          lx_val = lx_val - ld_stdate
        ENDIF
        REPLACE &lc_col WITH lx_val + &lc_col
        IF p_calc $ "STD,VAR"
          SELECT Squares
          GO recno( "crosstab" )
          REPLACE &lc_col WITH &lc_col + lx_val * lx_val
        ENDIF STD or VAR
      ELSE                              && p_calc $ "MAX,MIN"
        IF ln_count = 1
          REPLACE &lc_col WITH lx_val
        ELSE
          REPLACE &lc_col WITH &p_calc.( lx_val, &lc_col )
        ENDIF
      ENDIF
    ENDIF
    SELECT ( mainfile )

  ENDSCAN

**-------------------------------------------------------------------------
** We have finished with the data file.  We must now do a pass through Counts
** and Crosstab to adjust the values as required by PCT, AVG, STD or VAR.
** If CNT or PCT was requested, we simply copy the Counts file to Crosstab.
**
** If summaries are required, we compute them on the same pass into arrays.
** The arrays are used to summarize the rows of:
**      l_xarray   -- the columns of the Crosstab file
**      l_carray   -- the columns of the Counts file, holding counts
**      l_qarray   -- the columns of the Squares file, holding squares
** To save memory, element [1] of each array is used for column summaries,
** corresponding to the last ( ln_columns + 2 ) column of each file.
**-------------------------------------------------------------------------

  SELECT counts
  IF p_calc $ "CNT,PCT"                 && copy the counts in this case
    USE
    SELECT crosstab
    ZAP
    APPEND FROM XT_Count
    SELECT SELECT()
    USE XT_count ALIAS counts
  ENDIF

  IF p_summary                          && create and blank the arrays

    DO XT_BusyMsg WITH 5

    DECLARE l_xarray [ ln_columns + 1 ] && a variable per column of data
    DECLARE l_carray [ ln_columns + 1 ] && including the first
    IF p_calc $ "STD,VAR"
      DECLARE l_qarray[ ln_columns + 1 ]
    ENDIF
    ln_col = 1                           && first array column is "totals"
    DO WHILE ln_col <= ln_columns + 1
      DO CASE
        CASE lc_xtype $ "NF"
          STORE 0 TO l_xarray[ ln_col ]  && "blank" it
        CASE lc_xtype = "D"
          STORE {} TO l_xarray[ ln_col ]
        CASE lc_xtype = "C"
          STORE "" TO l_xarray[ ln_col ]
      ENDCASE
      STORE 0 TO l_carray[ ln_col ]
      IF p_calc $ "STD,VAR"
        STORE 0 TO l_qarray[ ln_col ]
      ENDIF
      ln_col = ln_col + 1
    ENDDO
  ENDIF

  GO TOP
  SCAN WHILE recno() <= ln_rows         && for each record in Counts

    ln_col = 2
    DO WHILE ln_col <= ln_columns + 1   && for each data field
      lc_col = field( ln_col )
      ln_num = &lc_col
      IF ln_num = 0                     && if no occurrences, loop
        ln_col = ln_col + 1
        LOOP
      ENDIF
      SELECT crosstab
      GO recno("counts")
      IF p_summary
        IF p_calc $ "MAX,MIN"
          IF l_carray [ln_col ] = 0
            l_xarray[ ln_col ] = &lc_col
          ELSE
            l_xarray[ ln_col ] = &p_calc.( l_xarray[ ln_col ], &lc_col )
          ENDIF
          IF l_carray[ 1 ] = 0
            l_xarray[ 1 ] = &lc_col
          ELSE
            l_xarray[ 1 ] = &p_calc.( l_xarray[ 1 ], &lc_col )
          ENDIF
        ELSE
          l_xarray[ ln_col ] = l_xarray[ ln_col ] + &lc_col
          l_xarray[ 1 ] = l_xarray[ 1 ] + &lc_col
        ENDIF
      ENDIF

      DO CASE
        CASE p_calc = "AVG"
          ln_num = &lc_col / ln_num
          REPLACE &lc_col WITH ln_num
        CASE p_calc $ "STD,VAR"
          SELECT squares
          GO recno("counts")
          ln_sq = &lc_col
          IF p_summary
            l_qarray[ ln_col ] = l_qarray[ ln_col ] + ln_sq
            l_qarray[ 1 ] =  l_qarray[ 1 ] + ln_sq
          ENDIF
          SELECT crosstab
          ln_var = ( ln_sq - &lc_col * &lc_col / ln_num ) ;
                    / ( ln_num - IIF( ll_sample, 1, 0 ) )
          REPLACE &lc_col WITH IIF( p_calc = "VAR", ln_var, sqrt( ln_var ) )
        CASE p_calc = "PCT"
          REPLACE &lc_col WITH &lc_col * 100 / ln_pop
      ENDCASE
      SELECT counts
      IF p_summary
        l_carray[ ln_col ] = l_carray[ ln_col ] + &lc_col
        l_carray[ 1 ] = l_carray[ 1 ] + &lc_col
      ENDIF
      ln_col = ln_col + 1
    ENDDO

    IF p_summary
      ** Do horizontal summary column
      lc_rtcol = field( ln_columns + 2 )
      SELECT counts
      REPLACE &lc_rtcol WITH l_carray[ 1 ]
      IF l_carray[ 1 ] > 0
        SELECT crosstab
        DO CASE
          CASE p_calc $ "CNT,SUM,MAX,MIN"
            REPLACE &lc_rtcol WITH l_xarray[ 1 ]
          CASE p_calc = "PCT"
            REPLACE &lc_rtcol WITH l_xarray[ 1 ] * 100 / ln_pop
          CASE p_calc = "AVG"
            REPLACE &lc_rtcol WITH l_xarray[ 1 ] / l_carray[ 1 ]
          OTHERWISE
            SELECT squares
            REPLACE &lc_rtcol WITH l_qarray[ 1 ]
            SELECT crosstab
            ln_var = ( l_qarray[ 1 ] - l_xarray[ 1 ] * l_xarray[ 1 ] ;
                 / l_carray[ 1 ] ) / ( l_carray[ 1 ] - IIF( ll_sample, 1, 0 ) )
            REPLACE &lc_rtcol WITH IIF(p_calc = "VAR", ln_var, SQRT( ln_var ) )
            STORE 0 TO l_qarray[ 1 ]
        ENDCASE
        SELECT counts
      ENDIF
      DO CASE
        CASE lc_xtype $ "NF"
          STORE 0 TO l_xarray[ 1 ]     && reinitialize at "blank"
        CASE lc_xtype = "D"
          STORE {} TO l_xarray[ 1 ]
        CASE lc_xtype = "C"
          STORE "" TO l_xarray[ 1 ]
      ENDCASE
      STORE 0 TO l_carray[ 1 ]
    ENDIF p_summary
  ENDSCAN

** now at bottom of crosstab data cells.  Write downward totals from arrays
** to the bottom records of the files and set up array elements [ 1 ] for
** grand totals
  SELECT crosstab
  GO BOTTOM
  SELECT counts
  GO BOTTOM
  IF p_summary
    ln_col = 2
    l_first = .T.
    DO WHILE ln_col <= ln_columns + 1
      lc_col = field( ln_col )
      ln_num = l_carray[ ln_col ]
      IF ln_num = 0                        && if no occurrences, loop
        ln_col = ln_col + 1
        LOOP
      ENDIF
      IF p_calc $ "MAX,MIN"
        IF l_first
          l_xarray[ 1 ] = l_xarray[ ln_col ]
          l_first = .F.
        ELSE
          l_xarray[ 1 ] = &p_calc.( l_xarray[ 1 ], l_xarray[ ln_col ] )
        ENDIF
      ELSE
        l_xarray[ 1 ] = l_xarray[ 1 ] + l_xarray[ ln_col ]
      ENDIF
      DO CASE
        CASE p_calc $ "STD,VAR"
          l_qarray[ 1 ] =  l_qarray[ 1 ] + l_qarray[ ln_col ]
          ln_var = ( l_qarray[ ln_col ] - l_xarray[ ln_col ] ;
                     * l_xarray[ ln_col ] / ln_num ) ;
                     / ( ln_num - IIF( ll_sample, 1, 0 ) )
          l_xarray[ ln_col ] = IIF( p_calc = "VAR", ln_var, sqrt( ln_var ) )
        CASE p_calc = "PCT"
          l_xarray[ ln_col ] = l_xarray[ ln_col ] * 100 / ln_pop
        CASE p_calc = "AVG"
          l_xarray[ ln_col ] = l_xarray[ ln_col ] / ln_num
      ENDCASE
      SELECT crosstab
      REPLACE &lc_col WITH l_xarray[ ln_col ]
      ln_col = ln_col + 1
      SELECT counts
      REPLACE &lc_col WITH ln_num
      l_carray[ 1 ] = l_carray[ 1 ] + ln_num
    ENDDO
    REPLACE &lc_rtcol WITH l_carray[ 1 ]

    ** Grand "total"
    SELECT crosstab
    DO CASE
      CASE p_calc $ "SUM,CNT,MAX,MIN"
        REPLACE &lc_rtcol WITH l_xarray[ 1 ]
      CASE p_calc = "PCT"
        REPLACE &lc_rtcol WITH l_xarray[ 1 ] * 100 / ln_pop
      CASE p_calc = "AVG"
        l_xarray[ 1 ] = l_xarray[ 1 ] / l_carray[ 1 ]
        REPLACE &lc_rtcol WITH l_xarray[ 1 ]
      OTHERWISE
        ln_var = ( l_qarray[ 1 ] - l_xarray[ 1 ] * l_xarray[ 1 ] / ;
                   l_carray[ 1 ] ) / ( l_carray[ 1 ] - IIF( ll_sample, 1, 0 ) )
        REPLACE &lc_rtcol WITH IIF(p_calc = "VAR", ln_var, sqrt(ln_var) )
    ENDCASE

  ENDIF p_summary

  DO XT_BusyMsg WITH 6

** create display file

  SET ORDER TO
  SELECT columns
  GO TOP
  SELECT select()
  USE _cross ALIAS cross
  GO TOP

  REPLACE field_name WITH lc_psidfld, ;
          field_type WITH "C", ;
          field_len  WITH max( ln_lensid, ln_lblmin ) , ;
          field_dec  WITH 0 , ;
          field_idx  WITH "N"

  SKIP
  SCAN WHILE recno() <=ln_columns + 1
    SELECT columns
    lc_coltitl = XT_ClnTl( &lc_ptopfld, lc_typetop, ln_xwid )
    SKIP
    SELECT cross
    REPLACE field_name WITH lc_coltitl, ;
            field_type WITH "C"       , ;
            field_len  WITH ln_xwid   , ;
            field_dec  WITH 0         , ;
            field_idx  WITH "N"

  ENDSCAN

  IF p_summary
    GO BOTTOM
    REPLACE field_name WITH "SUMMARY", ;
            field_type WITH "C"      , ;
            field_len  WITH ln_xwid  , ;
            field_dec  WITH 0        , ;
            field_idx  WITH "N"
  ENDIF

  CREATE XT_disp FROM _cross
  ERASE _cross.dbf

  DO XT_BusyMsg WITH 7

  SELECT crosstab
  SCAN
    SELECT XT_disp
    APPEND BLANK
    DO CASE
      CASE recno() > ln_rows
        REPLACE &lc_psidfld WITH "Summary"
      CASE lc_typesid = "D"
        lc_date = crosstab->&lc_psidfld
        lc_date = substr( lc_date, 5, 2 ) + "/" + right( lc_date, 2 ) ;
           + "/" + substr( lc_date, 3, 2 )
        REPLACE &lc_psidfld WITH lc_date
      OTHERWISE
        REPLACE &lc_psidfld WITH crosstab->&lc_psidfld
    ENDCASE
    ln_col = 2
    DO WHILE ln_col <= ln_columns + IIF( p_summary, 2, 1 )
      lc_col = field( ln_col )
      lc_xcol = field( ln_col, "crosstab" )
      SELECT counts
      GO recno( "crosstab" )
      SELECT XT_disp
      IF .NOT. ISBLANK( counts->&lc_xcol )              && skip empty cells
        DO CASE
          CASE lc_xtype = "C"
            lc_val = crosstab->&lc_xcol
            ** add code here to substitute "***BLANK***" or the like for
            ** any null strings, which may occur if there are meaningful
            ** blank character calculated fields in the original file and
            ** the operation is MIN.
          CASE lc_xtype = "D"
            lc_val = dtoc( crosstab->&lc_xcol )
          CASE lc_typecal = "D" .AND. p_calc = "AVG"
            lc_val = dtoc( crosstab->&lc_xcol + ld_stdate )
          OTHERWISE
            lc_val = str( crosstab->&lc_xcol, ln_xwid, ln_xdec )
        ENDCASE
        REPLACE &lc_col WITH lc_val
      ENDIF
      ln_col = ln_col + 1
    ENDDO
    SELECT crosstab
  ENDSCAN

  @ ln_disprow, 49 SAY "Complete"
  SELECT XT_disp
  DO XT_RestEv
RETURN
*-- EOP: Com_XTab

** This procedure added by JP
PROCEDURE XT_BusyMsg
PARAMETERS Message
*--------------------------------------------------------------------
* Print the progress message
* Parameter is the number of message
* routine uses global ( to it ) ln_disprow for the line number
*--------------------------------------------------------------------
  IF Message > 1
    @ ln_disprow, 49 SAY "Complete"
  ENDIF
  ln_disprow = ln_disprow + 1
  @ ln_disprow, 4 SAY lc_stat[ Message ] + CHR(205) + CHR(205) + CHR(16)
  @ ln_disprow, 49 SAY "Working"
RETURN
*-- EOP: XT_BusyMsg

PROCEDURE XT_IniVars && UDS - CrossTab()
*--------------------------------------------------------------------
* Declare the variables for the CrossTab form as PUBLIC and set
* their values.
*--------------------------------------------------------------------

  PUBLIC gc_xtitl,gc_xtop,gc_xside,gc_xsum,gc_xoper,gl_xtotals,gc_opmsg
  gc_xtitl  = SPACE(80)   && Title string for heading of Crosstab
  gc_xtop   = SPACE(80)   && Field for TOP, allow IIF expression
  gc_xside  = SPACE(80)   && Field for SIDE, allow IIF expression
  gc_xsum   = SPACE(80)   && Field to summarize, allow expressions
  gc_xoper  = SPACE(3)    && Operation to perform
  gl_xtotals= .F.         && Default is no summary row and column
  gc_opmsg  = ""

*-- Setup for Hand_Pick
  PUBLIC ARRAY aa_lookup[6]
  aa_lookup[1] = ""
  aa_lookup[2] = "FIELDS"
  aa_lookup[3] = "FIELDS"
  aa_lookup[4] = "FIELDS"
  aa_lookup[5] = ""
  aa_lookup[6] = ""

*-- Setup for Hand_Pick that specifies that the file is already open
  PUBLIC ARRAY params[3]
  params[1] = .T.                      && Lookup file open for this popup
  params[2] = ALIAS()                  && ALIAS of the open lookup file
  params[3] = ALIAS()                  && Current ALIAS that is in USE

RETURN
*-- EOP: XT_IniVars


FUNCTION XT_NavDsp  && UDS - CrossTab()
PARAMETER pn_bar
*-------------------------------------------------------------------
* Create the navigation line, and write it as a centered string
*-------------------------------------------------------------------
  PRIVATE lc_navline, ln_navcol

  IF ISBLANK( aa_lookup[ pn_bar ] )
    lc_navline = "Perform:F2  " +;
                 "Save only:Ctrl-End  " +;
                 "Cancel:Esc"
  ELSE
    lc_navline = "Pick operators/fields: Shift-F1  " +;
                 "Perform:F2  " +;
                 "Save only:Ctrl-End  " +;
                 "Cancel:Esc"
  ENDIF
  ln_navcol = Center( lc_navline, 80 )
  lc_win = WINDOW()
  ACTIVATE SCREEN
  SET CURSOR OFF
  IF "43" $ SET("DISPLAY")
    @ 41,0
    @ 41, ln_navcol SAY lc_navline COLOR &messcolor
  ELSE    
    @ 23,0
    @ 23, ln_navcol SAY lc_navline COLOR &messcolor
  ENDIF
  SET CURSOR ON
  ACTIVATE WINDOW &lc_win

RETURN( .T. )
*-- EOF: XT_NavDsp( pn_bar )

PROCEDURE XT_FldPick && UDS - CrossTab()
PARAMETERS pc_fldname
*-------------------------------------------------------------------
* Handle the pick list support for the Crosstab fields.  Actual
* picklist is handled by HAND_PICK in DB4LIB.PRG.
*-------------------------------------------------------------------

*-- Disable the Shift-F1 to prevent a recursive call
  ON KEY LABEL Shift-F1
  SET CURSOR OFF

*-- Get the field number for the aa_lookup array
*                        1         2         3         4         5
*               12345678901234567890123456789012345678901234567890
  lc_fldindx = "         GC_XTITL, GC_XTOP,  GC_XSIDE, GC_XSUM,  GC_XOPER"
  ln_fldno   = AT( pc_fldname, lc_fldindx ) / 10

  DO Hand_Pick WITH ln_fldno, .f.   && Call the pick list function

*-- Restore Shift-F1 back
  ON KEY LABEL Shift-F1 DO XT_FldPick WITH VARREAD()
  SET CURSOR ON

RETURN
*-- EOP: XT_FldPick WITH pc_FldName

FUNCTION XT_MakExt  && UDS - CrossTab()
*-------------------------------------------------------------------
* Make the EXX.PRG file if it does not exist.  Add it to the
* current catalog under the Application panel if it's not there.
* If the user is starting CrossTab() from the Dot prompt with no 
* active catalog, XT_MakExt will return a .F. value.
*-------------------------------------------------------------------
  PRIVATE lc_alias, ll_return, ln_catrec

  ll_return = .T.             && Assume the catalog is set up
  lc_alias = ALIAS()                                         

*-- Check to see if EXECCROS.PRG exists, if not create it in current dir
  IF .NOT. FILE( "EXX.PRG" )
    SET ALTERNATE TO exx.prg
    SET CONSOLE OFF
    SET ALTERNATE ON
    ?
    ?? "PROCEDURE Exx"        && Prevent blank line at top of file
    TEXT                      && Dump out the file
*------------------------------------------------------------------
* Execute the Crosstab routine based on the previous QBE with
* the CrossTab() function.
*
* Synopsis:
*
* DO EXX
*
* Description:
*
* EXX will use the current QBE view if the saved value
* in CROSSTAB.MEM (gc_xdbf) matches the currently opened DBF file.
* This will allow EXECCROS to use the correct FILTER if the QBE
* set one.  Otherwise, EXECCROS will close and open data files
* and re-open the DBF file CROSSTAB() used.
*
* Calls:
*
* COM_XTAB.PRG    Calculates the CROSSTAB.DBF based on any filter
*
* Files:
* 
* CROSSTAB.MEM    The memory file with the results from CROSSTAB()
* XT_DISP.DBF     The resulting display file from COM_XTAB.PRG
*
*------------------------------------------------------------------
  SET TALK OFF
  lc_catalog = SET( "CATALOG" )
  SET CATALOG OFF               
  ll_excl = SET( "EXCLUSIVE" ) = "ON"
  SET EXCLUSIVE ON

*-- Restore the values from the CrossTab() function form
  RESTORE FROM Crosstab ADDITIVE

**  Activate the file specified in Crosstab.mem
  lc_fullp = SET("FULLPATH")
  SET FULLPATH ON
  PRIVATE ln_alias
  ln_alias = 1
  DO WHILE ln_alias <= 40
    IF TRIM( DBF() ) = gc_xdbf
      EXIT
    ENDIF
    ln_alias = ln_alias + 1
  ENDDO
  SET FULLPATH &lc_fullp.
  IF ln_alias <= 40
    SELECT ( ln_alias )
  ELSE
    CLOSE DATABASES
    USE ( gc_xdbf )
  ENDIF

*-- Save the VIEW states for COM_XTAB to use
  gc_view   = SET( "VIEW" )

  SET FIELDS OFF                && Prevent problem of FIELDS not in view
  SET FIELDS TO                 && that were specified in CrossTab()
  lc_safety = SET("SAFETY")     && Set Safety off for overwriting files
  SET SAFETY OFF

*-- Perform the actual crosstab computations
  DO com_xtab WITH gc_xtop, gc_xside, gc_xsum, gc_xoper, gl_xtotals
  
*-- Display the results of the CrossTab in a BROWSE WINDOW with title
  CLEAR
  DEFINE WINDOW showxtab FROM 5,5 TO 15,76
  @ 3,0 SAY ""
  ?? gc_xtitl AT Center( gc_xtitl, 80 )
  SET DESIGN OFF                    && Prevent transfer to Query design
** Added to return to top of file
  GO TOP
  BROWSE WINDOW showxtab
  SET DESIGN ON
  @ 3,0
*-- Return the environment back the way we found it
  RELEASE ALL LIKE gc_x*
  RELEASE gn_len, gn_dec, lc_status, lc_escape, gl_xtotals
  RELEASE WINDOW showxtab
  SET CATALOG &lc_catalog.
  IF .NOT. ll_excl
    SET EXCLUSIVE OFF
  ENDIF
  SET SAFETY &lc_safety.      
RETURN
*-- EOP: ExecCros.prg
    ENDTEXT                   && End of the output for ExecCros.prg

    SET ALTERNATE OFF         && Close the output file
    SET CONSOLE ON
    SET ALTERNATE TO
  ENDIF

*?  SELECT ( lc_alias )       && Return the workarea back

RETURN( ll_return )
*-- EOF: XT_MakExt()         

PROCEDURE XT_RestEv && UDF - Com_XTab
*------------------------------------------------------------------
* Restore the previous environment
*------------------------------------------------------------------
  SET TALK &lc_talk
  SET SAFETY &lc_safety
  SET EXCLUSIVE &lc_excl
  SET STATUS &lc_status
  RELEASE WINDOWS com_xtab
  RELEASE gn_len, gn_dec, lc_status, lc_escape
RETURN
*-- EOP: XT_RestEv

FUNCTION XT_ClnTl   && UDS - Com_XTab
PARAMETERS p_title, p_type, p_width
*-----------------------------------------------------------------------
* Substantially revised by Jay Parsons, September, 1992
*
* Synopsis:
*
* XT_ClnTl( <p_title>, <p_type> )
*
* The XT_ClnTl() function takes a string parameter and returns a string
* acceptable as a fieldname for the XT_DISP file made from a value of that
* type.  Its second parameter represents the type of data being represented
* ( the type of the top field, not the type of the summary field ).
* It operates as follows:
*    If the <p_type> is "D", it assumes <p_title> is in DTOS format, YYYYMMDD.
* It converts a <p_title> of "19911025" to "OCT_25_91", but only if the result
* makes sense as a date.
*
*    If the <p_type> is "N" or "F", it adds leading zeroes as needed to make
* a 9-character string.
*
*    In the <p_type> is "L", it assumes the values are "F" or "T" and converts
* converts those to "FALSE" and "TRUE".
*
*    In all cases, if the result then contains any characters that are not valid
* in field names, it changes them to an underscore.  If the result then
* begins with a digit as converted numeric fields will, it prepends "F".
* It then truncates the result at 10 characters if required.
*
* Examples:
*
* XT_ClnTl("91.92","N") will return "F091_92"
* XT_Clntl("91.92","C") will return "F91_9"
* XT_ClnTl("AT&T","C" 4 ) will return "AT_T"
*
*-----------------------------------------------------------------------
  PRIVATE lc_result, ln_cnt, lc_valchars, lc_date, lc_mo, lc_day

  lc_valchars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"

  lc_result = upper( trim( ltrim ( p_title ) ) )

  DO CASE
    CASE p_type = "D" .AND. len( lc_result ) = 8
      IF val("1" + left( lc_result, 4 ) ) > 10000 && check for year all digits
        lc_mo = substr( lc_result, 5, 2 )
        lc_day = right( lc_result, 2 )
        lc_date = lc_mo +"/" + lc_day + "/" + left( lc_result, 4 )
        IF month( ctod( lc_date ) ) = val( lc_mo )
          lc_result = stuff( lc_date, 1, 2, ;
               substr("  JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC", ;
               3 * val( left( lc_date, 2 ) ), 3 ) )
          lc_result = left( lc_result, 7 ) + right( lc_result, 2 )
        ENDIF
      ENDIF
    CASE p_type $ "NF"
      lc_result = right( replicate( "0", 9 ) + lc_result, 9 )
    CASE p_type = "L" .AND. lc_result $ "TF"
      lc_result = IIF( lc_result = "F", "FALSE", "TRUE" )
  ENDCASE

*-- Loop through the characters and make the substitution as needed
  ln_cnt = 1
  DO WHILE ln_cnt <= len( lc_result )
    IF .NOT. substr( lc_result, ln_cnt, 1 ) $ lc_valchars
      lc_result = stuff( lc_result, ln_cnt, 1, "_" )
    ENDIF
    ln_cnt = ln_cnt + 1
  ENDDO

  IF .NOT. left( lc_result, 1 ) $ left( lc_valchars, 26 )
    lc_result = "F" + lc_result
  ENDIF

  lc_result = left( lc_result, 10 )
*
RETURN( lc_result )
*-- EOF: XT_ClnTl( <p_title>, <p_type>, <p_width> )



